"""Tests for the Bluetooth base scanner models."""

from __future__ import annotations

import asyncio
import time
from datetime import timedelta
from typing import Any
from unittest.mock import ANY, MagicMock, Mock, patch

import pytest
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from bleak_retry_connector import Allocations
from bluetooth_data_tools import monotonic_time_coarse

from habluetooth import (
    BaseHaRemoteScanner,
    BaseHaScanner,
    BluetoothManager,
    BluetoothScannerDevice,
    BluetoothScanningMode,
    HaBluetoothConnector,
    HaScannerDetails,
    HaScannerModeChange,
    HaScannerType,
    get_manager,
    set_manager,
)
from habluetooth.const import (
    CONNECTABLE_FALLBACK_MAXIMUM_STALE_ADVERTISEMENT_SECONDS,
    FALLBACK_MAXIMUM_STALE_ADVERTISEMENT_SECONDS,
    SCANNER_WATCHDOG_INTERVAL,
    SCANNER_WATCHDOG_TIMEOUT,
)
from habluetooth.storage import (
    DiscoveredDeviceAdvertisementData,
)

from . import (
    HCI0_SOURCE_ADDRESS,
    MockBleakClient,
    async_fire_time_changed,
    generate_advertisement_data,
    generate_ble_device,
    patch_bluetooth_time,
    utcnow,
)
from . import (
    InjectableRemoteScanner as FakeScanner,
)
from .conftest import FakeBluetoothAdapters, MockBluetoothManagerWithCallbacks


@pytest.mark.parametrize("name_2", [None, "w"])
@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_remote_scanner(name_2: str | None) -> None:
    """Test the remote scanner base class merges advertisement_data."""
    manager = get_manager()

    switchbot_device = generate_ble_device(
        "44:44:33:11:23:45",
        "wohand",
        {},
        rssi=-100,
    )
    switchbot_device_adv = generate_advertisement_data(
        local_name="wohand",
        service_uuids=["050a021a-0000-1000-8000-00805f9b34fb"],
        service_data={"050a021a-0000-1000-8000-00805f9b34fb": b"\n\xff"},
        manufacturer_data={1: b"\x01"},
        rssi=-100,
    )
    switchbot_device_2 = generate_ble_device(
        "44:44:33:11:23:45",
        name_2,
        {},
        rssi=-100,
    )
    switchbot_device_adv_2 = generate_advertisement_data(
        local_name=name_2,
        service_uuids=["00000001-0000-1000-8000-00805f9b34fb"],
        service_data={"00000001-0000-1000-8000-00805f9b34fb": b"\n\xff"},
        manufacturer_data={1: b"\x01", 2: b"\x02"},
        rssi=-100,
    )
    switchbot_device_3 = generate_ble_device(
        "44:44:33:11:23:45",
        "wohandlonger",
        {},
        rssi=-100,
    )
    switchbot_device_adv_3 = generate_advertisement_data(
        local_name="wohandlonger",
        service_uuids=["00000001-0000-1000-8000-00805f9b34fb"],
        service_data={"00000001-0000-1000-8000-00805f9b34fb": b"\n\xff"},
        manufacturer_data={1: b"\x01", 2: b"\x02"},
        rssi=-100,
    )
    switchbot_device_adv_4 = generate_advertisement_data(
        local_name="wohandlonger",
        service_uuids=["00000001-0000-1000-8000-00805f9b34fb"],
        service_data={"00000001-0000-1000-8000-00805f9b34fb": b"\n\xff"},
        manufacturer_data={1: b"\x04", 2: b"\x02", 3: b"\x03"},
        rssi=-100,
    )
    switchbot_device_adv_5 = generate_advertisement_data(
        local_name="wohandlonger",
        service_uuids=["00000001-0000-1000-8000-00805f9b34fb"],
        service_data={"00000001-0000-1000-8000-00805f9b34fb": b"\n\xff"},
        manufacturer_data={1: b"\x04", 2: b"\x01"},
        rssi=-100,
    )
    connector = HaBluetoothConnector(
        MockBleakClient, "mock_bleak_client", lambda: False
    )
    scanner = FakeScanner("esp32", "esp32", connector, True)
    details = scanner.details
    assert details == HaScannerDetails(
        source=scanner.source,
        connectable=scanner.connectable,
        name=scanner.name,
        adapter=scanner.adapter,
        scanner_type=HaScannerType.REMOTE,
    )
    unsetup = scanner.async_setup()
    cancel = manager.async_register_scanner(scanner)

    scanner.inject_advertisement(switchbot_device, switchbot_device_adv)

    data = scanner.discovered_devices_and_advertisement_data
    discovered_device, discovered_adv_data = data[switchbot_device.address]
    assert discovered_device.address == switchbot_device.address
    assert discovered_device.name == switchbot_device.name
    assert (
        discovered_adv_data.manufacturer_data == switchbot_device_adv.manufacturer_data
    )
    assert discovered_adv_data.service_data == switchbot_device_adv.service_data
    assert discovered_adv_data.service_uuids == switchbot_device_adv.service_uuids
    scanner.inject_advertisement(switchbot_device_2, switchbot_device_adv_2)

    data = scanner.discovered_devices_and_advertisement_data
    discovered_device, discovered_adv_data = data[switchbot_device.address]
    assert discovered_device.address == switchbot_device.address
    assert discovered_device.name == switchbot_device.name
    assert discovered_adv_data.manufacturer_data == {1: b"\x01", 2: b"\x02"}
    assert discovered_adv_data.service_data == {
        "050a021a-0000-1000-8000-00805f9b34fb": b"\n\xff",
        "00000001-0000-1000-8000-00805f9b34fb": b"\n\xff",
    }
    assert set(discovered_adv_data.service_uuids) == {
        "050a021a-0000-1000-8000-00805f9b34fb",
        "00000001-0000-1000-8000-00805f9b34fb",
    }

    # The longer name should be used
    scanner.inject_advertisement(switchbot_device_3, switchbot_device_adv_3)
    assert discovered_device.name == switchbot_device_3.name

    # Inject the shorter name / None again to make
    # sure we always keep the longer name
    scanner.inject_advertisement(switchbot_device_2, switchbot_device_adv_2)
    assert discovered_device.name == switchbot_device_3.name

    scanner.inject_advertisement(switchbot_device_2, switchbot_device_adv_4)
    assert scanner.discovered_devices_and_advertisement_data[
        switchbot_device_2.address
    ][1].manufacturer_data == {1: b"\x04", 2: b"\x02", 3: b"\x03"}
    scanner.inject_advertisement(switchbot_device_2, switchbot_device_adv_5)
    assert scanner.discovered_devices_and_advertisement_data[
        switchbot_device_2.address
    ][1].manufacturer_data == {1: b"\x04", 2: b"\x01", 3: b"\x03"}

    assert (
        "00090401-0052-036b-3206-ff0a050a021a"
        not in scanner.discovered_devices_and_advertisement_data[
            switchbot_device_2.address
        ][1].service_data
    )

    scanner.inject_raw_advertisement(
        switchbot_device_2.address,
        0,
        b"\x12\x21\x1a\x02\n\x05\n\xff\x062k\x03R\x00\x01\x04\t\x00\x04",
    )

    assert (
        "00090401-0052-036b-3206-ff0a050a021a"
        in scanner.discovered_devices_and_advertisement_data[
            switchbot_device_2.address
        ][1].service_data
    )

    assert scanner.serialize_discovered_devices() == DiscoveredDeviceAdvertisementData(
        connectable=True,
        expire_seconds=195,
        discovered_device_advertisement_datas={"44:44:33:11:23:45": ANY},
        discovered_device_timestamps={"44:44:33:11:23:45": ANY},
        discovered_device_raw={
            "44:44:33:11:23:45": b"\x12!\x1a\x02"
            b"\n\x05\n\xff"
            b"\x062k\x03"
            b"R\x00\x01\x04"
            b"\t\x00\x04"
        },
    )

    cancel()
    unsetup()


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_raw_advertisement_fast_path_unchanged() -> None:
    """Test that sending the same raw bytes twice uses the fast path."""
    manager = get_manager()

    connector = HaBluetoothConnector(
        MockBleakClient, "mock_bleak_client", lambda: False
    )
    scanner = FakeScanner("esp32", "esp32", connector, True)
    unsetup = scanner.async_setup()
    cancel = manager.async_register_scanner(scanner)

    raw_adv = b"\x12\x21\x1a\x02\n\x05\n\xff\x062k\x03R\x00\x01\x04\t\x00\x04"
    address = "44:44:33:11:23:45"

    # First raw advertisement — takes the slow path (parse + merge)
    scanner.inject_raw_advertisement(address, -60, raw_adv, 1.0)
    info1 = scanner._previous_service_info[address]
    assert info1.rssi == -60
    assert info1.raw == raw_adv

    # Second raw advertisement with same bytes — takes the fast path
    scanner.inject_raw_advertisement(address, -50, raw_adv, 2.0)
    info2 = scanner._previous_service_info[address]
    assert info2.rssi == -50
    assert info2.time == 2.0
    # Fast path reuses parsed data from prev_info
    assert info2.manufacturer_data is info1.manufacturer_data
    assert info2.service_data is info1.service_data
    assert info2.service_uuids is info1.service_uuids
    assert info2.name is info1.name
    assert info2.device is info1.device
    assert info2.raw is info1.raw

    # Third raw advertisement with different bytes — takes the slow path
    raw_adv_changed = b"\x12\x21\x1a\x02\n\x05\n\xff\x062k\x03R\x00\x01\x04\t\x00\x05"
    scanner.inject_raw_advertisement(address, -40, raw_adv_changed, 3.0)
    info3 = scanner._previous_service_info[address]
    assert info3.rssi == -40
    assert info3.raw == raw_adv_changed
    # Slow path re-parsed, so objects may differ
    assert info3.raw is not info1.raw

    cancel()
    unsetup()


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_remote_scanner_expires_connectable() -> None:
    """Test the remote scanner expires stale connectable data."""
    manager = get_manager()

    switchbot_device = generate_ble_device(
        "44:44:33:11:23:45",
        "wohand",
        {},
        rssi=-100,
    )
    switchbot_device_adv = generate_advertisement_data(
        local_name="wohand",
        service_uuids=[],
        manufacturer_data={1: b"\x01"},
        rssi=-100,
    )

    connector = HaBluetoothConnector(
        MockBleakClient, "mock_bleak_client", lambda: False
    )
    scanner = FakeScanner("esp32", "esp32", connector, True)
    unsetup = scanner.async_setup()
    cancel = manager.async_register_scanner(scanner)

    start_time_monotonic = time.monotonic()
    scanner.inject_advertisement(switchbot_device, switchbot_device_adv)

    devices = scanner.discovered_devices
    assert len(scanner.discovered_devices) == 1
    assert len(scanner.discovered_devices_and_advertisement_data) == 1
    assert devices[0].name == "wohand"

    expire_monotonic = (
        start_time_monotonic
        + CONNECTABLE_FALLBACK_MAXIMUM_STALE_ADVERTISEMENT_SECONDS
        + 1
    )
    expire_utc = utcnow() + timedelta(
        seconds=CONNECTABLE_FALLBACK_MAXIMUM_STALE_ADVERTISEMENT_SECONDS + 1
    )
    with patch_bluetooth_time(expire_monotonic):
        async_fire_time_changed(expire_utc)
        await asyncio.sleep(0)

    devices = scanner.discovered_devices
    assert len(scanner.discovered_devices) == 0
    assert len(scanner.discovered_devices_and_advertisement_data) == 0

    cancel()
    unsetup()


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_remote_scanner_expires_non_connectable() -> None:
    """Test the remote scanner expires stale non connectable data."""
    manager = get_manager()

    switchbot_device = generate_ble_device(
        "44:44:33:11:23:45",
        "wohand",
        {},
        rssi=-100,
    )
    switchbot_device_adv = generate_advertisement_data(
        local_name="wohand",
        service_uuids=[],
        manufacturer_data={1: b"\x01"},
        rssi=-100,
    )

    connector = HaBluetoothConnector(
        MockBleakClient, "mock_bleak_client", lambda: False
    )
    scanner = FakeScanner("esp32", "esp32", connector, True)
    unsetup = scanner.async_setup()
    cancel = manager.async_register_scanner(scanner)

    start_time_monotonic = time.monotonic()
    scanner.inject_advertisement(switchbot_device, switchbot_device_adv)

    devices = scanner.discovered_devices
    assert len(scanner.discovered_devices) == 1
    assert len(scanner.discovered_devices_and_advertisement_data) == 1
    assert len(scanner.discovered_device_timestamps) == 1
    dev_adv = scanner.get_discovered_device_advertisement_data(switchbot_device.address)
    assert dev_adv is not None
    dev, adv = dev_adv
    assert dev.name == "wohand"
    assert adv.local_name == "wohand"
    assert adv.manufacturer_data == switchbot_device_adv.manufacturer_data
    assert devices[0].name == "wohand"

    assert (
        FALLBACK_MAXIMUM_STALE_ADVERTISEMENT_SECONDS
        > CONNECTABLE_FALLBACK_MAXIMUM_STALE_ADVERTISEMENT_SECONDS
    )

    # The connectable timeout is used for all devices
    # as the manager takes care of availability and the scanner
    # if only concerned about making a connection
    expire_monotonic = (
        start_time_monotonic
        + CONNECTABLE_FALLBACK_MAXIMUM_STALE_ADVERTISEMENT_SECONDS
        + 1
    )
    expire_utc = utcnow() + timedelta(
        seconds=CONNECTABLE_FALLBACK_MAXIMUM_STALE_ADVERTISEMENT_SECONDS + 1
    )
    with patch_bluetooth_time(expire_monotonic):
        async_fire_time_changed(expire_utc)
        await asyncio.sleep(0)

    assert len(scanner.discovered_devices) == 0
    assert len(scanner.discovered_devices_and_advertisement_data) == 0
    assert len(scanner.discovered_device_timestamps) == 0
    assert (
        scanner.get_discovered_device_advertisement_data(switchbot_device.address)
        is None
    )

    expire_monotonic = (
        start_time_monotonic + FALLBACK_MAXIMUM_STALE_ADVERTISEMENT_SECONDS + 1
    )
    expire_utc = utcnow() + timedelta(
        seconds=FALLBACK_MAXIMUM_STALE_ADVERTISEMENT_SECONDS + 1
    )
    with patch_bluetooth_time(expire_monotonic):
        async_fire_time_changed(expire_utc)
        await asyncio.sleep(0)

    assert len(scanner.discovered_devices) == 0
    assert len(scanner.discovered_devices_and_advertisement_data) == 0
    assert len(scanner.discovered_device_timestamps) == 0
    assert (
        scanner.get_discovered_device_advertisement_data(switchbot_device.address)
        is None
    )

    cancel()
    unsetup()


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_remote_scanner_discovered_device_timestamps_deprecated() -> None:
    """The leading-underscore shim still returns timestamps but warns."""
    manager = get_manager()
    switchbot_device = generate_ble_device("44:44:33:11:23:45", "wohand", {}, rssi=-100)
    switchbot_device_adv = generate_advertisement_data(
        local_name="wohand", service_uuids=[], manufacturer_data={1: b"\x01"}, rssi=-100
    )
    connector = HaBluetoothConnector(
        MockBleakClient, "mock_bleak_client", lambda: False
    )
    scanner = FakeScanner("esp32", "esp32", connector, True)
    unsetup = scanner.async_setup()
    cancel = manager.async_register_scanner(scanner)
    scanner.inject_advertisement(switchbot_device, switchbot_device_adv)

    with pytest.warns(FutureWarning, match="_discovered_device_timestamps"):
        legacy = scanner._discovered_device_timestamps

    assert legacy == scanner.discovered_device_timestamps
    assert len(legacy) == 1

    cancel()
    unsetup()


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_base_scanner_connecting_behavior() -> None:
    """Test the default behavior is to mark the scanner as not scanning on connect."""
    manager = get_manager()

    switchbot_device = generate_ble_device(
        "44:44:33:11:23:45",
        "wohand",
        {},
        rssi=-100,
    )
    switchbot_device_adv = generate_advertisement_data(
        local_name="wohand",
        service_uuids=[],
        manufacturer_data={1: b"\x01"},
        rssi=-100,
    )

    connector = HaBluetoothConnector(
        MockBleakClient, "mock_bleak_client", lambda: False
    )
    scanner = FakeScanner("esp32", "esp32", connector, True)
    unsetup = scanner.async_setup()
    cancel = manager.async_register_scanner(scanner)

    with scanner.connecting():
        assert scanner.scanning is False

        # We should still accept new advertisements while connecting
        # since advertisements are delivered asynchronously and
        # we don't want to miss any even when we are willing to
        # accept advertisements from another scanner in the brief window
        # between when we start connecting and when we stop scanning
        scanner.inject_advertisement(switchbot_device, switchbot_device_adv)

    devices = scanner.discovered_devices
    assert len(scanner.discovered_devices) == 1
    assert len(scanner.discovered_devices_and_advertisement_data) == 1
    assert devices[0].name == "wohand"

    cancel()
    unsetup()


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_scanner_stops_responding() -> None:
    """Test we mark a scanner are not scanning when it stops responding."""
    manager = get_manager()

    connector = HaBluetoothConnector(
        MockBleakClient, "mock_bleak_client", lambda: False
    )
    scanner = FakeScanner(
        "esp32",
        "esp32",
        connector,
        True,
        current_mode=BluetoothScanningMode.ACTIVE,
        requested_mode=BluetoothScanningMode.ACTIVE,
    )
    unsetup = scanner.async_setup()
    cancel = manager.async_register_scanner(scanner)

    start_time_monotonic = time.monotonic()

    assert scanner.scanning is True
    failure_reached_time = (
        start_time_monotonic
        + SCANNER_WATCHDOG_TIMEOUT
        + SCANNER_WATCHDOG_INTERVAL.total_seconds()
    )
    # We hit the timer with no detections,
    # so we reset the adapter and restart the scanner
    with patch_bluetooth_time(failure_reached_time):
        async_fire_time_changed(utcnow() + SCANNER_WATCHDOG_INTERVAL)
        await asyncio.sleep(0)

    assert scanner.scanning is False

    bparasite_device = generate_ble_device(  # type: ignore[unreachable]
        "44:44:33:11:23:45",
        "bparasite",
        {},
        rssi=-100,
    )
    bparasite_device_adv = generate_advertisement_data(
        local_name="bparasite",
        service_uuids=[],
        manufacturer_data={1: b"\x01"},
        rssi=-100,
    )

    failure_reached_time += 1

    with patch_bluetooth_time(failure_reached_time):
        scanner.inject_advertisement(
            bparasite_device, bparasite_device_adv, failure_reached_time
        )

    # As soon as we get a detection, we know the scanner is working again
    assert scanner.scanning is True
    assert scanner.requested_mode == BluetoothScanningMode.ACTIVE
    assert scanner.current_mode == BluetoothScanningMode.ACTIVE

    cancel()
    unsetup()


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_merge_manufacturer_data_history_existing() -> None:
    """Test merging manufacturer data history."""
    manager = get_manager()

    sensor_push_device = generate_ble_device(
        "44:44:33:11:23:45",
        "",
        {},
        rssi=-60,
    )
    sensor_push_device_adv = generate_advertisement_data(
        local_name="",
        rssi=-60,
        manufacturer_data={
            64256: b"B\r.\xa9\xb6",
            31488: b"\x98\xfa\xb6\x91\xb6",
        },
        service_uuids=["ef090000-11d6-42ba-93b8-9dd7ec090ab0"],
        service_data={},
    )

    sensor_push_adv_2 = generate_advertisement_data(
        local_name="",
        service_uuids=["ef090000-11d6-42ba-93b8-9dd7ec090ab0"],
        service_data={},
        manufacturer_data={
            31488: b"\x98\xfa\xb6\x91\xb6",
        },
        rssi=-100,
    )

    connector = HaBluetoothConnector(
        MockBleakClient, "mock_bleak_client", lambda: False
    )
    scanner = FakeScanner("esp32", "esp32", connector, True)
    details = scanner.details
    assert details == HaScannerDetails(
        source=scanner.source,
        connectable=scanner.connectable,
        name=scanner.name,
        adapter=scanner.adapter,
        scanner_type=HaScannerType.REMOTE,
    )
    unsetup = scanner.async_setup()
    cancel = manager.async_register_scanner(scanner)

    scanner.inject_advertisement(sensor_push_device, sensor_push_device_adv)

    data = scanner.discovered_devices_and_advertisement_data
    discovered_device, discovered_adv_data = data[sensor_push_device.address]
    assert discovered_device.address == sensor_push_device.address
    assert discovered_device.name == sensor_push_device.name
    assert (
        discovered_adv_data.manufacturer_data
        == sensor_push_device_adv.manufacturer_data
    )
    assert discovered_adv_data.service_data == sensor_push_device_adv.service_data
    assert discovered_adv_data.service_uuids == sensor_push_device_adv.service_uuids
    scanner.inject_advertisement(sensor_push_device, sensor_push_adv_2)

    data = scanner.discovered_devices_and_advertisement_data
    discovered_device, discovered_adv_data = data[sensor_push_device.address]
    assert discovered_device.address == sensor_push_device.address
    assert discovered_device.name == sensor_push_device.name
    assert discovered_adv_data.manufacturer_data == {
        **sensor_push_device_adv.manufacturer_data,
        **sensor_push_adv_2.manufacturer_data,
    }
    assert discovered_adv_data.service_data == {}
    assert set(discovered_adv_data.service_uuids) == {
        *sensor_push_device_adv.service_uuids
    }

    cancel()
    unsetup()


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_merge_manufacturer_data_history_new() -> None:
    """Test merging manufacturer data history."""
    manager = get_manager()

    sensor_push_device = generate_ble_device(
        "44:44:33:11:23:45",
        "",
        {},
        rssi=-60,
    )
    sensor_push_device_adv = generate_advertisement_data(
        local_name="",
        rssi=-60,
        manufacturer_data={
            64256: b"B\r.\xa9\xb6",
            31488: b"\x98\xfa\xb6\x91\xb6",
        },
        service_uuids=["ef090000-11d6-42ba-93b8-9dd7ec090ab0"],
        service_data={},
    )

    sensor_push_adv_2 = generate_advertisement_data(
        local_name="",
        service_uuids=["ef090000-11d6-42ba-93b8-9dd7ec090ab0"],
        service_data={},
        manufacturer_data={
            21248: b"\xb9\xe9\xe1\xb9\xb6",
        },
        rssi=-100,
    )

    connector = HaBluetoothConnector(
        MockBleakClient, "mock_bleak_client", lambda: False
    )
    scanner = FakeScanner("esp32", "esp32", connector, True)
    details = scanner.details
    assert details == HaScannerDetails(
        source=scanner.source,
        connectable=scanner.connectable,
        name=scanner.name,
        adapter=scanner.adapter,
        scanner_type=HaScannerType.REMOTE,
    )
    unsetup = scanner.async_setup()
    cancel = manager.async_register_scanner(scanner)

    scanner.inject_advertisement(sensor_push_device, sensor_push_device_adv)

    data = scanner.discovered_devices_and_advertisement_data
    discovered_device, discovered_adv_data = data[sensor_push_device.address]
    assert discovered_device.address == sensor_push_device.address
    assert discovered_device.name == sensor_push_device.name
    assert (
        discovered_adv_data.manufacturer_data
        == sensor_push_device_adv.manufacturer_data
    )
    assert discovered_adv_data.service_data == sensor_push_device_adv.service_data
    assert discovered_adv_data.service_uuids == sensor_push_device_adv.service_uuids
    scanner.inject_advertisement(sensor_push_device, sensor_push_adv_2)

    data = scanner.discovered_devices_and_advertisement_data
    discovered_device, discovered_adv_data = data[sensor_push_device.address]
    assert discovered_device.address == sensor_push_device.address
    assert discovered_device.name == sensor_push_device.name
    assert discovered_adv_data.manufacturer_data == {
        **sensor_push_device_adv.manufacturer_data,
        **sensor_push_adv_2.manufacturer_data,
    }
    assert discovered_adv_data.service_data == {}
    assert set(discovered_adv_data.service_uuids) == {
        *sensor_push_device_adv.service_uuids
    }

    cancel()
    unsetup()


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_filter_apple_data() -> None:
    """Test filtering apple data accepts bytes that start with 01."""
    manager = get_manager()

    device = generate_ble_device(
        "44:44:33:11:23:45",
        "",
        {},
        rssi=-60,
    )
    device_adv = generate_advertisement_data(
        local_name="",
        rssi=-60,
        manufacturer_data={
            76: b"\x01\r.\xa9\xb6",
        },
        service_uuids=["ef090000-11d6-42ba-93b8-9dd7ec090ab0"],
        service_data={},
    )

    connector = HaBluetoothConnector(
        MockBleakClient, "mock_bleak_client", lambda: False
    )
    scanner = FakeScanner("esp32", "esp32", connector, True)
    details = scanner.details
    assert details == HaScannerDetails(
        source=scanner.source,
        connectable=scanner.connectable,
        name=scanner.name,
        adapter=scanner.adapter,
        scanner_type=HaScannerType.REMOTE,
    )
    unsetup = scanner.async_setup()
    cancel = manager.async_register_scanner(scanner)

    scanner.inject_advertisement(device, device_adv)

    data = scanner.discovered_devices_and_advertisement_data
    discovered_device, discovered_adv_data = data[device.address]
    assert discovered_device.address == device.address
    assert discovered_device.name == device.name
    assert discovered_adv_data.manufacturer_data == device_adv.manufacturer_data
    unsetup()
    cancel()


@pytest.mark.usefixtures("register_hci0_scanner")
def test_connection_history_count_in_progress() -> None:
    """Test connection history in process counting."""
    manager = get_manager()
    device1_address = "44:44:33:11:23:12"
    device2_address = "44:44:33:11:23:13"
    hci0_scanner = manager.async_scanner_by_source(HCI0_SOURCE_ADDRESS)
    assert hci0_scanner is not None
    hci0_scanner._add_connecting(device1_address)
    assert hci0_scanner._connections_in_progress() == 1
    hci0_scanner._add_connecting(device1_address)
    hci0_scanner._add_connecting(device2_address)
    assert hci0_scanner._connections_in_progress() == 3
    hci0_scanner._finished_connecting(device1_address, True)
    assert hci0_scanner._connections_in_progress() == 2
    hci0_scanner._finished_connecting(device1_address, False)
    assert hci0_scanner._connections_in_progress() == 1
    hci0_scanner._finished_connecting(device2_address, False)
    assert hci0_scanner._connections_in_progress() == 0


@pytest.mark.usefixtures("register_hci0_scanner")
def test_connection_history_failure_count(caplog: pytest.LogCaptureFixture) -> None:
    """Test connection history failure count."""
    manager = get_manager()
    device1_address = "44:44:33:11:23:12"
    device2_address = "44:44:33:11:23:13"
    hci0_scanner = manager.async_scanner_by_source(HCI0_SOURCE_ADDRESS)
    assert hci0_scanner is not None
    hci0_scanner._add_connecting(device1_address)
    hci0_scanner._finished_connecting(device1_address, False)
    assert hci0_scanner._connection_failures(device1_address) == 1
    hci0_scanner._add_connecting(device1_address)
    hci0_scanner._add_connecting(device2_address)
    hci0_scanner._finished_connecting(device1_address, False)
    assert hci0_scanner._connection_failures(device1_address) == 2
    hci0_scanner._finished_connecting(device2_address, False)
    assert hci0_scanner._connection_failures(device2_address) == 1
    hci0_scanner._add_connecting(device1_address)
    hci0_scanner._finished_connecting(device1_address, True)
    # On success, we should reset the failure count
    assert hci0_scanner._connection_failures(device1_address) == 0

    assert "Removing a non-existing connecting" not in caplog.text
    hci0_scanner._finished_connecting(device1_address, True)
    assert "Removing a non-existing connecting" in caplog.text


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_scanner_mode_changes() -> None:
    """Test scanner mode change methods notify the manager."""
    manager = get_manager()

    # Track mode changes
    mode_changes: list[HaScannerModeChange] = []

    def mode_callback(change: HaScannerModeChange) -> None:
        """Track mode changes."""
        mode_changes.append(change)

    cancel = manager.async_register_scanner_mode_change_callback(mode_callback, None)

    # Create a scanner with initial modes
    scanner = FakeScanner(
        HCI0_SOURCE_ADDRESS,
        "hci0",
        connectable=True,
        requested_mode=BluetoothScanningMode.PASSIVE,
        current_mode=BluetoothScanningMode.PASSIVE,
    )

    # Set up the scanner
    unsetup = scanner.async_setup()

    # Test changing requested mode
    scanner.set_requested_mode(BluetoothScanningMode.ACTIVE)
    assert len(mode_changes) == 1
    assert mode_changes[0].scanner == scanner
    assert mode_changes[0].requested_mode == BluetoothScanningMode.ACTIVE
    assert mode_changes[0].current_mode == BluetoothScanningMode.PASSIVE
    assert scanner.requested_mode == BluetoothScanningMode.ACTIVE

    # Test changing current mode
    scanner.set_current_mode(BluetoothScanningMode.ACTIVE)
    assert len(mode_changes) == 2
    assert mode_changes[1].scanner == scanner
    assert mode_changes[1].requested_mode == BluetoothScanningMode.ACTIVE
    assert mode_changes[1].current_mode == BluetoothScanningMode.ACTIVE
    assert scanner.current_mode == BluetoothScanningMode.ACTIVE

    # Test no notification when mode doesn't change
    scanner.set_current_mode(BluetoothScanningMode.ACTIVE)
    assert len(mode_changes) == 2  # No new notification

    # Test setting to None
    scanner.set_requested_mode(None)
    assert len(mode_changes) == 3
    assert mode_changes[2].requested_mode is None
    assert scanner.requested_mode is None

    scanner.set_current_mode(None)  # type: ignore[unreachable]
    assert len(mode_changes) == 4
    assert mode_changes[3].current_mode is None
    assert scanner.current_mode is None

    # Clean up
    unsetup()
    cancel()


def test_remote_scanner_type() -> None:
    """Test that remote scanners have REMOTE type."""

    class TestRemoteScanner(BaseHaRemoteScanner):
        """Test remote scanner implementation."""

    scanner = TestRemoteScanner("test_source", "test_adapter")
    assert scanner.details.scanner_type is HaScannerType.REMOTE


def test_base_scanner_with_connector() -> None:
    """Test BaseHaScanner with connector and adapter type."""
    manager = get_manager()

    mock_adapters: dict[str, dict[str, Any]] = {
        "test_adapter": {
            "address": "00:1A:7D:DA:71:04",
            "adapter_type": "usb",
        }
    }

    connector = HaBluetoothConnector(
        client=MagicMock, source="test_source", can_connect=lambda: True
    )

    with patch.object(manager, "_adapters", mock_adapters):
        scanner = BaseHaScanner(
            source="test_source",
            adapter="test_adapter",
            connector=connector,
            connectable=True,
        )
        assert scanner.details.scanner_type is HaScannerType.USB


class TestScanner(BaseHaScanner):
    """Test scanner without slots for mocking."""

    __test__ = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._test_allocations = None

    def get_allocations(self):
        """Override to return test allocations."""
        return self._test_allocations


def test_score_with_no_allocations():
    """Test scoring when no allocation info is available."""
    scanner = TestScanner(
        source="test_source",
        adapter="test_adapter",
        connectable=True,
    )

    ble_device = BLEDevice(
        address="00:11:22:33:44:55",
        name="Test Device",
        details={},
    )
    advertisement = AdvertisementData(
        local_name="Test Device",
        manufacturer_data={},
        service_data={},
        service_uuids=[],
        rssi=-50,
        platform_data=(),
        tx_power=None,
    )
    scanner_device = BluetoothScannerDevice(
        scanner=scanner,
        ble_device=ble_device,
        advertisement=advertisement,
    )

    # No allocation info available, should just use RSSI
    score = scanner._score_connection_paths(10, scanner_device)
    assert score == -50


def test_score_with_all_slots_free():
    """Test scoring when all slots are free."""
    scanner = TestScanner(
        source="test_source",
        adapter="test_adapter",
        connectable=True,
    )

    # Set test allocations to return all slots free
    scanner._test_allocations = Allocations(
        adapter="test_adapter",
        slots=5,
        free=5,
        allocated=[],
    )

    ble_device = BLEDevice(
        address="00:11:22:33:44:55",
        name="Test Device",
        details={},
    )
    advertisement = AdvertisementData(
        local_name="Test Device",
        manufacturer_data={},
        service_data={},
        service_uuids=[],
        rssi=-50,
        platform_data=(),
        tx_power=None,
    )
    scanner_device = BluetoothScannerDevice(
        scanner=scanner,
        ble_device=ble_device,
        advertisement=advertisement,
    )

    # All slots free, no penalty
    score = scanner._score_connection_paths(10, scanner_device)
    assert score == -50


def test_score_with_one_slot_remaining():
    """Test scoring when only one slot remains."""
    scanner = TestScanner(
        source="test_source",
        adapter="test_adapter",
        connectable=True,
    )

    # Set test allocations to return only 1 slot free
    scanner._test_allocations = Allocations(
        adapter="test_adapter",
        slots=5,
        free=1,
        allocated=[
            "AA:BB:CC:DD:EE:01",
            "AA:BB:CC:DD:EE:02",
            "AA:BB:CC:DD:EE:03",
            "AA:BB:CC:DD:EE:04",
        ],
    )

    ble_device = BLEDevice(
        address="00:11:22:33:44:55",
        name="Test Device",
        details={},
    )
    advertisement = AdvertisementData(
        local_name="Test Device",
        manufacturer_data={},
        service_data={},
        service_uuids=[],
        rssi=-50,
        platform_data=(),
        tx_power=None,
    )
    scanner_device = BluetoothScannerDevice(
        scanner=scanner,
        ble_device=ble_device,
        advertisement=advertisement,
    )

    # One slot remaining, small penalty (rssi_diff * 0.76)
    score = scanner._score_connection_paths(10, scanner_device)
    assert score == -50 - (10 * 0.76)
    assert score == -57.6


def test_score_with_no_slots_available():
    """Test scoring when no slots are available."""
    scanner = TestScanner(
        source="test_source",
        adapter="test_adapter",
        connectable=True,
    )

    # Set test allocations to return no slots free
    scanner._test_allocations = Allocations(
        adapter="test_adapter",
        slots=5,
        free=0,
        allocated=[
            "AA:BB:CC:DD:EE:01",
            "AA:BB:CC:DD:EE:02",
            "AA:BB:CC:DD:EE:03",
            "AA:BB:CC:DD:EE:04",
            "AA:BB:CC:DD:EE:05",
        ],
    )

    ble_device = BLEDevice(
        address="00:11:22:33:44:55",
        name="Test Device",
        details={},
    )
    advertisement = AdvertisementData(
        local_name="Test Device",
        manufacturer_data={},
        service_data={},
        service_uuids=[],
        rssi=-50,
        platform_data=(),
        tx_power=None,
    )
    scanner_device = BluetoothScannerDevice(
        scanner=scanner,
        ble_device=ble_device,
        advertisement=advertisement,
    )

    # No slots available, returns NO_RSSI_VALUE (-127)
    score = scanner._score_connection_paths(10, scanner_device)
    assert score == -127


def test_score_comparison_with_different_slot_availability():
    """Test that scanners with more free slots score better."""
    # Scanner with all slots free
    scanner_all_free = TestScanner(
        source="adapter1",
        adapter="adapter1",
        connectable=True,
    )

    # Scanner with one slot remaining
    scanner_one_slot = TestScanner(
        source="adapter2",
        adapter="adapter2",
        connectable=True,
    )

    # Scanner with no slots
    scanner_no_slots = TestScanner(
        source="adapter3",
        adapter="adapter3",
        connectable=True,
    )

    ble_device = BLEDevice(
        address="00:11:22:33:44:55",
        name="Test Device",
        details={},
    )
    advertisement = AdvertisementData(
        local_name="Test Device",
        manufacturer_data={},
        service_data={},
        service_uuids=[],
        rssi=-50,
        platform_data=(),
        tx_power=None,
    )

    # Create scanner devices for each scanner
    device_all_free = BluetoothScannerDevice(
        scanner=scanner_all_free,
        ble_device=ble_device,
        advertisement=advertisement,
    )
    device_one_slot = BluetoothScannerDevice(
        scanner=scanner_one_slot,
        ble_device=ble_device,
        advertisement=advertisement,
    )
    device_no_slots = BluetoothScannerDevice(
        scanner=scanner_no_slots,
        ble_device=ble_device,
        advertisement=advertisement,
    )

    # Set allocations for each scanner
    scanner_all_free._test_allocations = Allocations(
        adapter="adapter1",
        slots=5,
        free=5,
        allocated=[],
    )
    scanner_one_slot._test_allocations = Allocations(
        adapter="adapter2",
        slots=5,
        free=1,
        allocated=[
            "AA:BB:CC:DD:EE:01",
            "AA:BB:CC:DD:EE:02",
            "AA:BB:CC:DD:EE:03",
            "AA:BB:CC:DD:EE:04",
        ],
    )
    scanner_no_slots._test_allocations = Allocations(
        adapter="adapter3",
        slots=5,
        free=0,
        allocated=[
            "AA:BB:CC:DD:EE:01",
            "AA:BB:CC:DD:EE:02",
            "AA:BB:CC:DD:EE:03",
            "AA:BB:CC:DD:EE:04",
            "AA:BB:CC:DD:EE:05",
        ],
    )

    score_all_free = scanner_all_free._score_connection_paths(10, device_all_free)
    score_one_slot = scanner_one_slot._score_connection_paths(10, device_one_slot)
    score_no_slots = scanner_no_slots._score_connection_paths(10, device_no_slots)

    # Verify the scoring order: all_free > one_slot > no_slots
    assert score_all_free > score_one_slot
    assert score_one_slot > score_no_slots

    # Verify specific values
    assert score_all_free == -50
    assert score_one_slot == -57.6
    assert score_no_slots == -127  # NO_RSSI_VALUE


def test_score_with_connections_in_progress_and_slots():
    """Test that both connection progress and slot availability are considered."""
    scanner = TestScanner(
        source="test_source",
        adapter="test_adapter",
        connectable=True,
    )

    # Add a connection in progress
    scanner._add_connecting("FF:EE:DD:CC:BB:AA")

    # Set test allocations to return only 1 slot free
    scanner._test_allocations = Allocations(
        adapter="test_adapter",
        slots=5,
        free=1,
        allocated=[
            "AA:BB:CC:DD:EE:01",
            "AA:BB:CC:DD:EE:02",
            "AA:BB:CC:DD:EE:03",
            "AA:BB:CC:DD:EE:04",
        ],
    )

    ble_device = BLEDevice(
        address="00:11:22:33:44:55",
        name="Test Device",
        details={},
    )
    advertisement = AdvertisementData(
        local_name="Test Device",
        manufacturer_data={},
        service_data={},
        service_uuids=[],
        rssi=-50,
        platform_data=(),
        tx_power=None,
    )
    scanner_device = BluetoothScannerDevice(
        scanner=scanner,
        ble_device=ble_device,
        advertisement=advertisement,
    )

    # Both penalties should apply
    score = scanner._score_connection_paths(10, scanner_device)
    # -50 (RSSI) - 10.1 (connection in progress) - 7.6 (last slot)
    assert score == -50 - 10.1 - 7.6
    assert score == -67.7


@pytest.mark.asyncio
async def test_on_scanner_start_callback_remote_scanner(
    async_mock_manager_with_scanner_callbacks: MockBluetoothManagerWithCallbacks,
) -> None:
    """Test that on_scanner_start is called when a remote scanner starts."""
    manager = async_mock_manager_with_scanner_callbacks

    # Create a fake remote scanner
    scanner = FakeScanner(
        source="esp32_proxy",
        adapter="esp32_proxy",
        connector=None,
        connectable=True,
    )

    # Simulate scanner start success
    scanner._on_start_success()

    # Verify the callback was called
    assert len(manager.scanner_start_calls) == 1
    assert manager.scanner_start_calls[0] is scanner


@pytest.mark.asyncio
async def test_on_scanner_start_multiple_scanners(
    async_mock_manager_with_scanner_callbacks: MockBluetoothManagerWithCallbacks,
) -> None:
    """Test that on_scanner_start is called for multiple scanners."""
    manager = async_mock_manager_with_scanner_callbacks

    # Create multiple scanners
    scanner1 = FakeScanner(
        source="scanner1",
        adapter="scanner1",
        connector=None,
        connectable=True,
    )

    scanner2 = FakeScanner(
        source="scanner2",
        adapter="scanner2",
        connector=None,
        connectable=True,
    )

    # Simulate both scanners starting
    scanner1._on_start_success()
    scanner2._on_start_success()

    # Verify both callbacks were called
    assert len(manager.scanner_start_calls) == 2
    assert scanner1 in manager.scanner_start_calls
    assert scanner2 in manager.scanner_start_calls


@pytest.mark.asyncio
async def test_scanner_without_manager() -> None:
    """Test that _on_start_success handles scanner without manager gracefully."""
    # Set a temporary manager for scanner creation
    mock_bluetooth_adapters = FakeBluetoothAdapters()
    temp_manager = BluetoothManager(
        mock_bluetooth_adapters,
        slot_manager=Mock(),
    )

    original_manager = get_manager()
    set_manager(temp_manager)

    try:
        # Create a scanner
        scanner = FakeScanner(
            source="test",
            adapter="test",
            connector=None,
            connectable=True,
        )

        # Clear the manager to simulate no manager scenario
        scanner._manager = None  # type: ignore[assignment]

        # Should not raise an exception
        scanner._on_start_success()

    finally:
        # Restore original manager
        set_manager(original_manager)


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_remote_scanner_restore_discovered_devices() -> None:
    """Test serialize→restore round-trip on a BaseHaRemoteScanner."""
    manager = get_manager()
    connector = HaBluetoothConnector(
        MockBleakClient, "mock_bleak_client", lambda: False
    )

    # First scanner: receives advertisements, then serializes.
    src_scanner = FakeScanner("esp32", "esp32", connector, True)
    src_unsetup = src_scanner.async_setup()
    src_cancel = manager.async_register_scanner(src_scanner)

    address = "44:44:33:11:23:45"
    device = generate_ble_device(address, "wohand", {})
    adv = generate_advertisement_data(
        local_name="wohand",
        service_uuids=["050a021a-0000-1000-8000-00805f9b34fb"],
        service_data={"050a021a-0000-1000-8000-00805f9b34fb": b"\n\xff"},
        manufacturer_data={1: b"\x01"},
        rssi=-60,
        tx_power=4,
    )
    now = monotonic_time_coarse()
    src_scanner.inject_advertisement(device, adv, now=now)
    raw_adv = b"\x12\x21\x1a\x02\n\x05\n\xff\x062k\x03R\x00\x01\x04\t\x00\x04"
    src_scanner.inject_raw_advertisement(address, -55, raw_adv, now=now + 1.0)

    history = src_scanner.serialize_discovered_devices()
    assert history.discovered_device_raw[address] == raw_adv
    assert history.discovered_device_timestamps[address] == now + 1.0

    # Second scanner: fresh instance, restore from serialized blob.
    dst_scanner = FakeScanner("esp32-replay", "esp32-replay", connector, True)
    dst_unsetup = dst_scanner.async_setup()
    dst_cancel = manager.async_register_scanner(dst_scanner)

    assert dst_scanner.discovered_devices == []
    dst_scanner.restore_discovered_devices(history)

    restored = dst_scanner.discovered_devices_and_advertisement_data
    assert set(restored) == {address}
    rest_device, rest_adv = restored[address]
    assert rest_device.address == address
    assert rest_device.name == "wohand"
    assert rest_adv.manufacturer_data == adv.manufacturer_data
    assert rest_adv.service_uuids == adv.service_uuids
    # Restored source overrides the previous scanner's source — the device
    # belongs to the scanner that owns it now.
    assert dst_scanner._previous_service_info[address].source == "esp32-replay"
    # Raw advertisement bytes survive the round-trip.
    assert dst_scanner._previous_service_info[address].raw == raw_adv

    src_cancel()
    src_unsetup()
    dst_cancel()
    dst_unsetup()


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_remote_scanner_async_diagnostics() -> None:
    """Test BaseHaRemoteScanner.async_diagnostics shape."""
    manager = get_manager()
    connector = HaBluetoothConnector(
        MockBleakClient, "mock_bleak_client", lambda: False
    )
    scanner = FakeScanner("esp32", "esp32", connector, True)
    unsetup = scanner.async_setup()
    cancel = manager.async_register_scanner(scanner)

    address = "44:44:33:11:23:45"
    device = generate_ble_device(address, "wohand", {})
    adv = generate_advertisement_data(
        local_name="wohand",
        service_uuids=[],
        service_data={},
        manufacturer_data={1: b"\x01"},
        rssi=-60,
    )
    now = monotonic_time_coarse()
    scanner.inject_advertisement(device, adv, now=now)
    raw = b"\x02\x01\x06"
    scanner.inject_raw_advertisement(address, -55, raw, now=now + 1.0)

    diagnostics = await scanner.async_diagnostics()
    # Remote scanner adds these three keys on top of the base diagnostics.
    assert diagnostics["raw_advertisement_data"] == {address: raw}
    assert diagnostics["discovered_device_timestamps"] == {address: now + 1.0}
    assert set(diagnostics["time_since_last_device_detection"]) == {address}
    # Base keys still present.
    assert diagnostics["source"] == "esp32"
    assert diagnostics["connectable"] is True
    # Connection counters start empty.
    assert diagnostics["connect_in_progress"] == {}
    assert diagnostics["connect_failures"] == {}

    cancel()
    unsetup()


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_diagnostics_surface_connect_state() -> None:
    """async_diagnostics should expose connect-in-progress and failure counts."""
    manager = get_manager()
    scanner = FakeScanner(
        source="diag", adapter="diag", connector=None, connectable=True
    )
    unsetup = scanner.async_setup()
    cancel = manager.async_register_scanner(scanner)

    addr_a = "AA:BB:CC:DD:EE:01"
    addr_b = "AA:BB:CC:DD:EE:02"

    # Two attempts in flight for addr_a, one for addr_b.
    scanner._add_connecting(addr_a)
    scanner._add_connecting(addr_a)
    scanner._add_connecting(addr_b)

    # addr_a then succeeds (clears its failures), addr_b fails twice.
    scanner._finished_connecting(addr_a, connected=True)
    scanner._finished_connecting(addr_b, connected=False)
    scanner._add_connecting(addr_b)
    scanner._finished_connecting(addr_b, connected=False)

    diagnostics = await scanner.async_diagnostics()
    # addr_a still has one in-flight attempt, addr_b is fully drained.
    assert diagnostics["connect_in_progress"] == {addr_a: 1}
    assert diagnostics["connect_failures"] == {addr_b: 2}
    # Lifetime counters: one success (addr_a), two failures (addr_b twice).
    assert diagnostics["connect_completed_total"] == 1
    assert diagnostics["connect_failed_total"] == 2
    assert diagnostics["last_connect_completed_time"] > 0.0
    # Returned dicts must be copies — mutating them doesn't poison scanner state.
    diagnostics["connect_in_progress"].clear()
    diagnostics["connect_failures"].clear()
    assert scanner._connect_in_progress == {addr_a: 1}
    assert scanner._connect_failures == {addr_b: 2}

    cancel()
    unsetup()


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_lifetime_connect_counters_reset_on_history_clear() -> None:
    """Lifetime counters should reset whenever connection history is cleared."""
    manager = get_manager()
    scanner = FakeScanner(
        source="counters", adapter="counters", connector=None, connectable=True
    )
    unsetup = scanner.async_setup()
    cancel = manager.async_register_scanner(scanner)

    addr = "AA:BB:CC:DD:EE:99"
    # Three successes, one failure → totals reflect lifetime, not per-address state.
    for _ in range(3):
        scanner._add_connecting(addr)
        scanner._finished_connecting(addr, connected=True)
    scanner._add_connecting(addr)
    scanner._finished_connecting(addr, connected=False)

    assert scanner._connect_completed_total == 3
    assert scanner._connect_failed_total == 1
    assert scanner._last_connect_completed_time > 0.0
    # Per-address failure count survives because success preceded the last failure.
    assert scanner._connect_failures == {addr: 1}

    scanner._clear_connection_history()
    assert scanner._connect_completed_total == 0
    assert scanner._connect_failed_total == 0
    assert scanner._last_connect_completed_time == 0.0
    assert scanner._connect_failures == {}
    assert scanner._connect_in_progress == {}

    cancel()
    unsetup()


@pytest.mark.usefixtures("enable_bluetooth")
@pytest.mark.asyncio
async def test_set_mode_noop_when_unchanged() -> None:
    """set_requested_mode and set_current_mode should be no-ops on same value."""
    manager = get_manager()
    scanner = FakeScanner(
        source="modes",
        adapter="modes",
        connector=None,
        connectable=True,
        requested_mode=BluetoothScanningMode.ACTIVE,
    )
    with patch.object(manager, "scanner_mode_changed") as notify:
        # Same value → setter is a no-op, no notify.
        scanner.set_requested_mode(BluetoothScanningMode.ACTIVE)
        notify.assert_not_called()

        # Different value → notify fires once.
        scanner.set_requested_mode(BluetoothScanningMode.PASSIVE)
        assert notify.call_count == 1

        # current_mode was None at init; setting None is a no-op.
        scanner.set_current_mode(None)
        assert notify.call_count == 1

        # Different value → notify fires again.
        scanner.set_current_mode(BluetoothScanningMode.PASSIVE)
        assert notify.call_count == 2

        # Repeating same value is a no-op.
        scanner.set_current_mode(BluetoothScanningMode.PASSIVE)
        assert notify.call_count == 2



"""Serialize/Deserialize bluetooth adapter discoveries."""

from __future__ import annotations

import logging
import time
from dataclasses import dataclass, field
from typing import Any, Final, TypedDict

from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

_LOGGER = logging.getLogger(__name__)


@dataclass
class DiscoveredDeviceAdvertisementData:
    """Discovered device advertisement data deserialized from storage."""

    connectable: bool
    expire_seconds: float
    discovered_device_advertisement_datas: dict[
        str, tuple[BLEDevice, AdvertisementData]
    ]
    discovered_device_timestamps: dict[str, float]
    discovered_device_raw: dict[str, bytes | None] = field(default_factory=dict)


CONNECTABLE: Final = "connectable"
EXPIRE_SECONDS: Final = "expire_seconds"
DISCOVERED_DEVICE_ADVERTISEMENT_DATAS: Final = "discovered_device_advertisement_datas"
DISCOVERED_DEVICE_TIMESTAMPS: Final = "discovered_device_timestamps"
DISCOVERED_DEVICE_RAW: Final = "discovered_device_raw"


class DiscoveredDeviceAdvertisementDataDict(TypedDict):
    """Discovered device advertisement data dict in storage."""

    connectable: bool
    expire_seconds: float
    discovered_device_advertisement_datas: dict[str, DiscoveredDeviceDict]
    discovered_device_timestamps: dict[str, float]
    discovered_device_raw: dict[str, str | None]


ADDRESS: Final = "address"
NAME: Final = "name"
RSSI: Final = "rssi"
DETAILS: Final = "details"


class BLEDeviceDict(TypedDict):
    """BLEDevice dict."""

    address: str
    name: str | None
    rssi: int | None  # Kept for backward compatibility
    details: dict[str, Any]


LOCAL_NAME: Final = "local_name"
MANUFACTURER_DATA: Final = "manufacturer_data"
SERVICE_DATA: Final = "service_data"
SERVICE_UUIDS: Final = "service_uuids"
TX_POWER: Final = "tx_power"
PLATFORM_DATA: Final = "platform_data"


class AdvertisementDataDict(TypedDict):
    """AdvertisementData dict."""

    local_name: str | None
    manufacturer_data: dict[str, str]
    service_data: dict[str, str]
    service_uuids: list[str]
    rssi: int
    tx_power: int | None
    platform_data: list[Any]


class DiscoveredDeviceDict(TypedDict):
    """Discovered device dict."""

    device: BLEDeviceDict
    advertisement_data: AdvertisementDataDict


def expire_stale_scanner_discovered_device_advertisement_data(
    data_by_scanner: dict[str, DiscoveredDeviceAdvertisementDataDict],
) -> None:
    """Expire stale discovered device advertisement data."""
    now = time.time()
    expired_scanners: list[str] = []
    for scanner, data in data_by_scanner.items():
        expire: list[str] = []
        expire_seconds = data[EXPIRE_SECONDS]
        timestamps = data[DISCOVERED_DEVICE_TIMESTAMPS]
        discovered_device_advertisement_datas = data[
            DISCOVERED_DEVICE_ADVERTISEMENT_DATAS
        ]
        discovered_device_raw = data.get(DISCOVERED_DEVICE_RAW, {})
        for address, timestamp in timestamps.items():
            time_diff = now - timestamp
            if time_diff > expire_seconds:
                expire.append(address)
            elif time_diff < 0:
                _LOGGER.warning(
                    "Discarding timestamp %s for %s on "
                    "scanner %s as it is the future (now = %s)",
                    timestamp,
                    address,
                    scanner,
                    now,
                )
                expire.append(address)
        for address in expire:
            del timestamps[address]
            del discovered_device_advertisement_datas[address]
            discovered_device_raw.pop(address, None)
        if not timestamps:
            expired_scanners.append(scanner)
        _LOGGER.debug(
            "Loaded %s fresh discovered devices for %s", len(timestamps), scanner
        )

    for scanner in expired_scanners:
        del data_by_scanner[scanner]


def discovered_device_advertisement_data_from_dict(
    data: DiscoveredDeviceAdvertisementDataDict,
) -> DiscoveredDeviceAdvertisementData | None:
    """Build discovered_device_advertisement_data dict."""
    try:
        return DiscoveredDeviceAdvertisementData(
            data[CONNECTABLE],
            data[EXPIRE_SECONDS],
            _deserialize_discovered_device_advertisement_datas(
                data[DISCOVERED_DEVICE_ADVERTISEMENT_DATAS]
            ),
            _deserialize_discovered_device_timestamps(
                data[DISCOVERED_DEVICE_TIMESTAMPS]
            ),
            _deserialize_discovered_device_raw(data.get(DISCOVERED_DEVICE_RAW, {})),
        )
    except (KeyError, ValueError, TypeError):
        _LOGGER.warning(
            "Discovery cache shape mismatch, discarding cache; "
            "adapter startup will be slow"
        )
    except Exception:  # pylint: disable=broad-except
        _LOGGER.exception(
            "Unexpected error deserializing discovered_device_advertisement_data, "
            "adapter startup will be slow"
        )
    return None


def discovered_device_advertisement_data_to_dict(
    data: DiscoveredDeviceAdvertisementData,
) -> DiscoveredDeviceAdvertisementDataDict:
    """Build discovered_device_advertisement_data dict."""
    return DiscoveredDeviceAdvertisementDataDict(
        connectable=data.connectable,
        expire_seconds=data.expire_seconds,
        discovered_device_advertisement_datas=_serialize_discovered_device_advertisement_datas(
            data.discovered_device_advertisement_datas
        ),
        discovered_device_timestamps=_serialize_discovered_device_timestamps(
            data.discovered_device_timestamps
        ),
        discovered_device_raw=_serialize_discovered_device_raw(
            data.discovered_device_raw
        ),
    )


def _serialize_discovered_device_advertisement_datas(
    discovered_device_advertisement_datas: dict[
        str, tuple[BLEDevice, AdvertisementData]
    ],
) -> dict[str, DiscoveredDeviceDict]:
    """Serialize discovered_device_advertisement_datas."""
    return {
        address: DiscoveredDeviceDict(
            device=_ble_device_to_dict(device, advertisement_data),
            advertisement_data=_advertisement_data_to_dict(advertisement_data),
        )
        for (
            address,
            (device, advertisement_data),
        ) in discovered_device_advertisement_datas.items()
    }


def _deserialize_discovered_device_advertisement_datas(
    discovered_device_advertisement_datas: dict[str, DiscoveredDeviceDict],
) -> dict[str, tuple[BLEDevice, AdvertisementData]]:
    """Deserialize discovered_device_advertisement_datas."""
    return {
        address: (
            _ble_device_from_dict(device_advertisement_data["device"]),
            _advertisement_data_from_dict(
                device_advertisement_data["advertisement_data"]
            ),
        )
        for (
            address,
            device_advertisement_data,
        ) in discovered_device_advertisement_datas.items()
    }


def _ble_device_from_dict(device_dict: BLEDeviceDict) -> BLEDevice:
    """Deserialize BLEDevice from dict, handling backward compatibility."""
    # Remove rssi from dict as BLEDevice no longer accepts it in bleak 1.x
    device_data = device_dict.copy()
    device_data.pop("rssi", None)  # type: ignore[misc]  # Remove rssi if present (backward compatibility)
    return BLEDevice(**device_data)


def _ble_device_to_dict(
    ble_device: BLEDevice, advertisement_data: AdvertisementData
) -> BLEDeviceDict:
    """Serialize ble_device."""
    return BLEDeviceDict(
        address=ble_device.address,
        name=ble_device.name,
        rssi=advertisement_data.rssi,  # For backward compatibility
        details=ble_device.details,
    )


def _advertisement_data_from_dict(
    advertisement_data: AdvertisementDataDict,
) -> AdvertisementData:
    """Deserialize advertisement_data."""
    return AdvertisementData(
        local_name=advertisement_data[LOCAL_NAME],
        manufacturer_data={
            int(manufacturer_id): bytes.fromhex(manufacturer_data)
            for manufacturer_id, manufacturer_data in advertisement_data[
                MANUFACTURER_DATA
            ].items()
        },
        service_data={
            service_uuid: bytes.fromhex(service_data)
            for service_uuid, service_data in advertisement_data[SERVICE_DATA].items()
        },
        service_uuids=advertisement_data[SERVICE_UUIDS],
        rssi=advertisement_data[RSSI],
        tx_power=advertisement_data[TX_POWER],
        platform_data=tuple(advertisement_data[PLATFORM_DATA]),
    )


def _advertisement_data_to_dict(
    advertisement_data: AdvertisementData,
) -> AdvertisementDataDict:
    """Serialize advertisement_data."""
    return AdvertisementDataDict(
        local_name=advertisement_data.local_name,
        manufacturer_data={
            str(manufacturer_id): manufacturer_data.hex()
            for manufacturer_id, manufacturer_data in advertisement_data.manufacturer_data.items()  # noqa: E501
        },
        service_data={
            service_uuid: service_data.hex()
            for service_uuid, service_data in advertisement_data.service_data.items()
        },
        service_uuids=advertisement_data.service_uuids,
        rssi=advertisement_data.rssi,
        tx_power=advertisement_data.tx_power,
        platform_data=list(advertisement_data.platform_data),
    )


def _get_monotonic_time_diff() -> float:
    """Get monotonic time diff."""
    return time.time() - time.monotonic()


def _deserialize_discovered_device_timestamps(
    discovered_device_timestamps: dict[str, float],
) -> dict[str, float]:
    """Deserialize discovered_device_timestamps."""
    time_diff = _get_monotonic_time_diff()
    return {
        address: unix_time - time_diff
        for address, unix_time in discovered_device_timestamps.items()
    }


def _serialize_discovered_device_timestamps(
    discovered_device_timestamps: dict[str, float],
) -> dict[str, float]:
    """Serialize discovered_device_timestamps."""
    time_diff = _get_monotonic_time_diff()
    return {
        address: monotonic_time + time_diff
        for address, monotonic_time in discovered_device_timestamps.items()
    }


def _deserialize_discovered_device_raw(
    discovered_device_raw: dict[str, str | None],
) -> dict[str, bytes | None]:
    """Deserialize discovered_device_timestamps."""
    return {
        address: None if raw is None else bytes.fromhex(raw)
        for address, raw in discovered_device_raw.items()
    }


def _serialize_discovered_device_raw(
    discovered_device_raw: dict[str, bytes | None],
) -> dict[str, str | None]:
    """Serialize discovered_device_timestamps."""
    return {
        address: None if raw is None else raw.hex()
        for address, raw in discovered_device_raw.items()
    }


DiscoveryStorageType = dict[str, DiscoveredDeviceAdvertisementDataDict]



from __future__ import annotations

import asyncio
import logging
from asyncio import timeout as asyncio_timeout
from contextlib import asynccontextmanager
from struct import Struct
from typing import TYPE_CHECKING, cast

from btsocket import btmgmt_socket
from btsocket.btmgmt_socket import BluetoothSocketError

from ..const import (
    FAST_CONN_LATENCY,
    FAST_CONN_TIMEOUT,
    FAST_MAX_CONN_INTERVAL,
    FAST_MIN_CONN_INTERVAL,
    MEDIUM_CONN_LATENCY,
    MEDIUM_CONN_TIMEOUT,
    MEDIUM_MAX_CONN_INTERVAL,
    MEDIUM_MIN_CONN_INTERVAL,
    ConnectParams,
)

if TYPE_CHECKING:
    import socket
    from collections.abc import AsyncIterator, Callable

    from ..scanner import HaScanner

_LOGGER = logging.getLogger(__name__)
_int = int
_bytes = bytes
# Everything is little endian

HEADER_SIZE = 6
# Header is event_code (2 bytes), controller_idx (2 bytes), param_len (2 bytes)
DEVICE_FOUND = 0x0012
ADV_MONITOR_DEVICE_FOUND = 0x002F

# Management commands
MGMT_OP_GET_CONNECTIONS = 0x0015
MGMT_OP_LOAD_CONN_PARAM = 0x0035

# Management events
MGMT_EV_CMD_COMPLETE = 0x0001
MGMT_EV_CMD_STATUS = 0x0002

# Pre-compiled struct formats for performance
COMMAND_HEADER = Struct("<HHH")
COMMAND_HEADER_PACK = COMMAND_HEADER.pack
CONN_PARAM_STRUCT = Struct("<H6sBHHHH")
CONN_PARAM_PACK = CONN_PARAM_STRUCT.pack

CONNECTION_ERRORS = (
    BluetoothSocketError,
    OSError,
    asyncio.TimeoutError,
    PermissionError,
)


def _set_future_if_not_done(future: asyncio.Future[None] | None) -> None:
    """Set the future result if not done."""
    if future is not None and not future.done():
        future.set_result(None)


class BluetoothMGMTProtocol:
    """Bluetooth MGMT protocol."""

    def __init__(
        self,
        connection_made_future: asyncio.Future[None],
        scanners: dict[int, HaScanner],
        on_connection_lost: Callable[[], None],
        is_shutting_down: Callable[[], bool],
        sock: socket.socket,
    ) -> None:
        """Initialize the protocol."""
        self.transport: asyncio.Transport | None = None
        self.connection_made_future = connection_made_future
        self._buffer: bytes | None = None
        self._buffer_len = 0
        self._pos = 0
        self._scanners = scanners
        self._on_connection_lost = on_connection_lost
        self._is_shutting_down = is_shutting_down
        self._pending_commands: dict[int, asyncio.Future[tuple[int, bytes]]] = {}
        self._sock = sock

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Handle connection made."""
        _set_future_if_not_done(self.connection_made_future)
        self.transport = cast("asyncio.Transport", transport)

    def _write_to_socket(self, data: bytes) -> None:
        """
        Write data directly to the socket, bypassing asyncio transport.

        This works around a kernel bug where sendto() on Bluetooth management
        sockets returns 0 instead of the number of bytes sent on some platforms
        (e.g., Odroid M1 with kernel 6.12.43). When asyncio sees 0, it thinks
        the send failed and retries forever.

        Since mgmt sockets are SOCK_RAW, sends are atomic - either the entire
        packet is sent or nothing is sent.
        """
        try:
            n = self._sock.send(data)
            # On buggy kernels, n might be 0 even though the data was sent
            # We treat 0 as success for mgmt sockets
            if n == 0 and len(data) > 0:
                # Kernel bug: returned 0 but data was actually sent
                _LOGGER.debug(
                    "Bluetooth mgmt socket returned 0 for %d bytes (kernel bug fix)",
                    len(data),
                )
        except Exception:
            _LOGGER.exception("Failed to write to mgmt socket")
            raise

    @asynccontextmanager
    async def command_response(
        self, opcode: int
    ) -> AsyncIterator[asyncio.Future[tuple[int, bytes]]]:
        """
        Context manager for handling command responses.

        Usage:
            async with protocol.command_response(opcode) as future:
                transport.write(command)
                status, data = await future
        """
        future: asyncio.Future[tuple[int, bytes]] = (
            asyncio.get_running_loop().create_future()
        )
        self._pending_commands[opcode] = future
        try:
            yield future
        finally:
            # Clean up if the future wasn't resolved
            self._pending_commands.pop(opcode, None)

    def _add_to_buffer(self, data: bytes | bytearray | memoryview) -> None:
        """Add data to the buffer."""
        # Protractor sends a bytearray, so we need to convert it to bytes
        # https://github.com/esphome/issues/issues/5117
        # type(data) should not be isinstance(data, bytes) because we want to
        # to explicitly check for bytes and not for subclasses of bytes
        bytes_data = bytes(data) if type(data) is not bytes else data
        if self._buffer_len == 0:
            # This is the best case scenario, we don't have to copy the data
            # and can just use the buffer directly. This is the most common
            # case as well.
            self._buffer = bytes_data
        else:
            if TYPE_CHECKING:
                assert self._buffer is not None, "Buffer should be set"
            # This is the worst case scenario, we have to copy the bytes_data
            # and can't just use the buffer directly. This is also very
            # uncommon since we usually read the entire frame at once.
            self._buffer += bytes_data
        self._buffer_len += len(bytes_data)

    def _remove_from_buffer(self) -> None:
        """Remove data from the buffer."""
        end_of_frame_pos = self._pos
        self._buffer_len -= end_of_frame_pos
        if self._buffer_len == 0:
            # This is the best case scenario, we can just set the buffer to None
            # and don't have to copy the data. This is the most common case as well.
            self._buffer = None
            return
        if TYPE_CHECKING:
            assert self._buffer is not None, "Buffer should be set"
        # This is the worst case scenario, we have to copy the data
        # and can't just use the buffer directly. This should only happen
        # when we read multiple frames at once because the event loop
        # is blocked and we cannot pull the data out of the buffer fast enough.
        cstr = self._buffer
        # Important: we must use the explicit length for the slice
        # since Cython will stop at any '\0' character if we don't
        self._buffer = cstr[end_of_frame_pos : self._buffer_len + end_of_frame_pos]

    def data_received(self, data: _bytes) -> None:  # noqa: C901
        """Handle data received."""
        self._add_to_buffer(data)
        while self._buffer_len >= 6:
            if TYPE_CHECKING:
                assert self._buffer is not None, "Buffer should be set"
            self._pos = 6
            header = self._buffer
            event_code = header[0] | (header[1] << 8)
            controller_idx = header[2] | (header[3] << 8)
            param_len = header[4] | (header[5] << 8)
            if self._buffer_len < self._pos + param_len:
                # We don't have the entire frame yet, so we need to wait
                # for more data to arrive.
                return
            self._pos += param_len
            if event_code == DEVICE_FOUND:
                parse_offset = 6
            elif event_code == ADV_MONITOR_DEVICE_FOUND:
                parse_offset = 8
            elif event_code in {MGMT_EV_CMD_COMPLETE, MGMT_EV_CMD_STATUS}:
                # Handle management command responses
                if param_len >= 3:
                    opcode = header[6] | (header[7] << 8)
                    status = header[8]
                    if opcode == MGMT_OP_LOAD_CONN_PARAM:
                        self._handle_load_conn_param_response(status, controller_idx)
                    elif (
                        opcode == MGMT_OP_GET_CONNECTIONS
                        and opcode in self._pending_commands
                    ):
                        # Handle GET_CONNECTIONS response for capability check
                        future = self._pending_commands.pop(opcode)
                        if not future.done():
                            # Return status and any response data
                            response_data = (
                                header[9 : self._pos] if param_len > 3 else b""
                            )
                            future.set_result((status, response_data))
                self._remove_from_buffer()
                continue
            else:
                self._remove_from_buffer()
                continue
            address = header[parse_offset : parse_offset + 6]
            address_type = header[parse_offset + 6]
            rssi = header[parse_offset + 7]
            if rssi > 128:
                rssi -= 256

            flags = (
                header[parse_offset + 8]
                | (header[parse_offset + 9] << 8)
                | (header[parse_offset + 10] << 16)
                | (header[parse_offset + 11] << 24)
            )

            # Skip AD_Data_Length (2 bytes) at parse_offset+12 and +13
            data = header[parse_offset + 14 : self._pos]
            self._remove_from_buffer()

            if (scanner := self._scanners.get(controller_idx)) is not None:
                # We have a scanner for this controller, so we can
                # pass the data to it.
                scanner._async_on_raw_bluez_advertisement(
                    address,
                    address_type,
                    rssi,
                    flags,
                    data,
                )

    def _handle_load_conn_param_response(
        self, status: _int, controller_idx: _int
    ) -> None:
        """Handle MGMT_OP_LOAD_CONN_PARAM response."""
        if status != 0:
            _LOGGER.warning(
                "hci%u: Failed to load conn params: status=%d",
                controller_idx,
                status,
            )
        else:
            _LOGGER.debug(
                "hci%u: Connection parameters loaded successfully",
                controller_idx,
            )

    def connection_lost(self, exc: Exception | None) -> None:
        """Handle connection lost."""
        # Only suppress warnings during shutdown, not info messages
        if exc:
            if not self._is_shutting_down():
                _LOGGER.warning("Bluetooth management socket connection lost: %s", exc)
        else:
            _LOGGER.info("Bluetooth management socket connection closed")
        self.transport = None
        self._on_connection_lost()


class MGMTBluetoothCtl:
    """Class to control interfaces using the BlueZ management API."""

    def __init__(self, timeout: float, scanners: dict[int, HaScanner]) -> None:
        """Initialize the control class."""
        # Internal state
        self.timeout = timeout
        self.protocol: BluetoothMGMTProtocol | None = None
        self.sock: socket.socket | None = None
        self.scanners = scanners
        self._reconnect_task: asyncio.Task[None] | None = None
        self._on_connection_lost_future: asyncio.Future[None] | None = None
        self._shutting_down = False

    def close(self) -> None:
        """Close the management interface."""
        self._shutting_down = True
        if self._reconnect_task:
            self._reconnect_task.cancel()
        if self.protocol and self.protocol.transport:
            self.protocol.transport.close()
            self.protocol = None
        btmgmt_socket.close(self.sock)

    def _on_connection_lost(self) -> None:
        """Handle connection lost."""
        if self._shutting_down:
            _LOGGER.debug("Bluetooth management socket connection lost during shutdown")
        else:
            _LOGGER.debug("Bluetooth management socket connection lost, reconnecting")
        _set_future_if_not_done(self._on_connection_lost_future)
        self._on_connection_lost_future = None

    async def reconnect_task(self) -> None:
        """Monitor the connection and reconnect if needed."""
        while not self._shutting_down:
            if self._on_connection_lost_future:
                await self._on_connection_lost_future
            if self._shutting_down:
                break  # type: ignore[unreachable]
            _LOGGER.debug("Reconnecting to Bluetooth management socket")
            try:
                await self._establish_connection()
            except CONNECTION_ERRORS:
                _LOGGER.debug("Bluetooth management socket connection timed out")
                # If we get a timeout, we should try to reconnect
                # after a short delay
                await asyncio.sleep(1)

    async def _establish_connection(self) -> None:
        """Establish a connection to the Bluetooth management socket."""
        _LOGGER.debug("Establishing Bluetooth management socket connection")
        self.sock = btmgmt_socket.open()
        loop = asyncio.get_running_loop()
        connection_made_future: asyncio.Future[None] = loop.create_future()
        try:
            async with asyncio_timeout(self.timeout):
                # _create_connection_transport accessed
                # directly to avoid SOCK_STREAM check
                # see https://bugs.python.org/issue38285
                _, protocol = await loop._create_connection_transport(  # type: ignore[attr-defined]
                    self.sock,
                    lambda: BluetoothMGMTProtocol(
                        connection_made_future,
                        self.scanners,
                        self._on_connection_lost,
                        lambda: self._shutting_down,
                        self.sock,
                    ),
                    None,
                    None,
                )
                await connection_made_future
        except TimeoutError:
            btmgmt_socket.close(self.sock)
            raise
        _LOGGER.debug("Bluetooth management socket connection established")
        self.protocol = cast("BluetoothMGMTProtocol", protocol)
        self._on_connection_lost_future = loop.create_future()

    def _has_mgmt_capabilities_from_status(self, status: int) -> bool:
        """
        Check if a MGMT command status indicates we have capabilities.

        Returns True if we have capabilities, False otherwise.

        Status codes:
        - 0x00 = Success (we have permissions)
        - 0x01 = Unknown Command (might happen if kernel is too old)
        - 0x0D = Invalid Parameters
        - 0x10 = Not Powered (for some operations)
        - 0x11 = Invalid Index (adapter doesn't exist but we have permissions)
        - 0x14 = Permission Denied (missing NET_ADMIN/NET_RAW)
        """
        if status == 0x14:  # Permission denied
            _LOGGER.debug(
                "MGMT capability check failed with permission denied - "
                "missing NET_ADMIN/NET_RAW"
            )
            return False
        if status in (0x00, 0x11):  # Success or Invalid Index
            _LOGGER.debug("MGMT capability check passed (status: %#x)", status)
            return True
        # Unknown status - log it and assume no permissions to be safe
        _LOGGER.debug(
            "MGMT capability check returned unexpected status %#x - "
            "assuming missing permissions",
            status,
        )
        return False

    async def _check_capabilities(self) -> bool:
        """
        Check if we have the necessary capabilities to use MGMT.

        Returns True if we have capabilities, False otherwise.
        """
        if not self.protocol or not self.protocol.transport:
            return False

        # Try GET_CONNECTIONS for adapter 0 - this is a read-only command
        # that requires NET_ADMIN privileges but doesn't change any state
        header = COMMAND_HEADER_PACK(
            MGMT_OP_GET_CONNECTIONS,  # opcode
            0,  # controller index 0 (hci0)
            0,  # no parameters
        )

        try:
            return await self._do_mgmt_op_get_connections(header)
        except (TimeoutError, OSError) as ex:
            _LOGGER.debug(
                "MGMT capability check failed: %s - likely missing NET_ADMIN/NET_RAW",
                ex,
            )
            return False

    async def _do_mgmt_op_get_connections(self, header: bytes) -> bool:
        """Send a MGMT_OP_GET_CONNECTIONS command and check capabilities."""
        if TYPE_CHECKING:
            assert self.protocol is not None
            assert self.protocol.transport is not None

        async with self.protocol.command_response(
            MGMT_OP_GET_CONNECTIONS
        ) as response_future:
            self.protocol._write_to_socket(header)
            # Wait for response with timeout
            async with asyncio_timeout(5.0):
                status, _ = await response_future
            return self._has_mgmt_capabilities_from_status(status)

    async def setup(self) -> None:
        """Set up management interface."""
        await self._establish_connection()

        # Check if we actually have the capabilities to use MGMT
        if not await self._check_capabilities():
            # Mark as shutting down to prevent reconnection attempts
            self._shutting_down = True
            # Close the connection and raise an error to trigger fallback
            if self.protocol and self.protocol.transport:
                self.protocol.transport.close()
            btmgmt_socket.close(self.sock)
            msg = "Missing NET_ADMIN/NET_RAW capabilities for Bluetooth management"
            raise PermissionError(msg)

        self._reconnect_task = asyncio.create_task(self.reconnect_task())

    def load_conn_params(
        self,
        adapter_idx: int,
        address: str,
        address_type: int,
        params: ConnectParams,
    ) -> bool:
        """
        Load connection parameters for a specific device.

        Args:
            adapter_idx: Adapter index (e.g., 0 for hci0)
            address: Device MAC address (e.g., "AA:BB:CC:DD:EE:FF")
            address_type: BDADDR_LE_PUBLIC (1) or BDADDR_LE_RANDOM (2)
            params: Connection parameters to load (ConnectParams.FAST or
              ConnectParams.MEDIUM)

        Returns:
            True if command was sent successfully

        """
        if not self.protocol or not self.protocol.transport:
            _LOGGER.error("Cannot load conn params: no connection")
            return False

        # Parse MAC address
        addr_bytes = bytes.fromhex(address.replace(":", ""))
        if len(addr_bytes) != 6:
            _LOGGER.error("Invalid MAC address: %s", address)
            return False

        # Build command structure (C definitions from BlueZ mgmt-api.txt):
        # struct mgmt_cp_load_conn_param {
        #     uint16_t param_count;
        #     struct mgmt_conn_param params[0];
        # };
        # struct mgmt_conn_param {
        #     struct mgmt_addr_info addr;
        #     uint16_t min_interval;
        #     uint16_t max_interval;
        #     uint16_t latency;
        #     uint16_t timeout;
        # };
        # struct mgmt_addr_info {
        #     bdaddr_t bdaddr;
        #     uint8_t type;
        # };

        # Get the appropriate connection parameters based on the enum
        if params is ConnectParams.FAST:
            min_interval = FAST_MIN_CONN_INTERVAL
            max_interval = FAST_MAX_CONN_INTERVAL
            latency = FAST_CONN_LATENCY
            timeout = FAST_CONN_TIMEOUT
        else:  # params is ConnectParams.MEDIUM:
            min_interval = MEDIUM_MIN_CONN_INTERVAL
            max_interval = MEDIUM_MAX_CONN_INTERVAL
            latency = MEDIUM_CONN_LATENCY
            timeout = MEDIUM_CONN_TIMEOUT

        # Pack the command
        cmd_data = CONN_PARAM_PACK(
            1,  # param_count = 1
            addr_bytes[::-1],  # bdaddr (reversed for little endian)
            address_type,  # address type
            min_interval,  # min_interval
            max_interval,  # max_interval
            latency,  # latency
            timeout,  # timeout
        )

        # Send the command
        try:
            header = COMMAND_HEADER_PACK(
                MGMT_OP_LOAD_CONN_PARAM,  # opcode
                adapter_idx,  # controller index
                len(cmd_data),  # parameter length
            )
            self.protocol._write_to_socket(header + cmd_data)
            _LOGGER.debug(
                "Loaded conn params for %s: interval=%d-%d, latency=%d, timeout=%d",
                address,
                min_interval,
                max_interval,
                latency,
                timeout,
            )
        except Exception:
            _LOGGER.exception("Failed to load conn params")
            return False
        else:
            return True

    def load_conn_params_explicit(
        self,
        adapter_idx: int,
        address: str,
        address_type: int,
        min_interval: int,
        max_interval: int,
        latency: int,
        timeout: int,
    ) -> bool:
        """
        Load explicit connection parameters for a specific device.

        Args:
            adapter_idx: Adapter index (e.g., 0 for hci0)
            address: Device MAC address (e.g., "AA:BB:CC:DD:EE:FF")
            address_type: BDADDR_LE_PUBLIC (1) or BDADDR_LE_RANDOM (2)
            min_interval: Minimum connection interval (units of 1.25ms)
            max_interval: Maximum connection interval (units of 1.25ms)
            latency: Connection latency (number of events)
            timeout: Supervision timeout (units of 10ms)

        Returns:
            True if command was sent successfully

        """
        if not self.protocol or not self.protocol.transport:
            _LOGGER.error("Cannot load conn params: no connection")
            return False

        # Parse MAC address
        addr_bytes = bytes.fromhex(address.replace(":", ""))
        if len(addr_bytes) != 6:
            _LOGGER.error("Invalid MAC address: %s", address)
            return False

        # Pack the command
        cmd_data = CONN_PARAM_PACK(
            1,  # param_count = 1
            addr_bytes[::-1],  # bdaddr (reversed for little endian)
            address_type,  # address type
            min_interval,
            max_interval,
            latency,
            timeout,
        )

        # Send the command
        try:
            header = COMMAND_HEADER_PACK(
                MGMT_OP_LOAD_CONN_PARAM,  # opcode
                adapter_idx,  # controller index
                len(cmd_data),  # parameter length
            )
            self.protocol._write_to_socket(header + cmd_data)
            _LOGGER.debug(
                "Loaded explicit conn params for %s:"
                " interval=%d-%d, latency=%d, timeout=%d",
                address,
                min_interval,
                max_interval,
                latency,
                timeout,
            )
        except Exception:
            _LOGGER.exception("Failed to load conn params")
            return False
        else:
            return True
